# w.c.s. - web application for online forms
# Copyright (C) 2005-2018  Entr'ouvert
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

from quixote import get_publisher, get_request, get_response, get_session, redirect
from quixote.directory import Directory
from quixote.errors import PublishError

from wcs.carddef import CardDef
from wcs.formdef import FormDef
from wcs.forms.common import FormTemplateMixin
from wcs.qommon import _, errors, misc, template
from wcs.qommon.afterjobs import AfterJob
from wcs.qommon.backoffice.menu import html_top as backoffice_html_top
from wcs.qommon.form import Form
from wcs.qommon.http_request import HTTPRequest
from wcs.wf.jump import jump_and_perform
from wcs.workflows import perform_items


class MissingOrExpiredToken(PublishError):
    status_code = 404
    title = _('Error')
    description = _('This action link has already been used or has expired.')


class MissingFormdata(PublishError):
    status_code = 404
    title = _('Error')
    description = _('This action link is no longer valid as the attached form has been removed.')


class ActionsDirectory(Directory):
    def _q_lookup(self, component):
        try:
            token = get_publisher().token_class.get(component)
        except KeyError:
            raise MissingOrExpiredToken()
        if token.type == 'action':
            return ActionDirectory(token)
        if token.type == 'global-interactive-action':
            return GlobalInteractiveActionDirectory(token)
        raise errors.TraversalError()


class ActionDirectory(Directory, FormTemplateMixin):
    _q_exports = ['']
    templates = ['wcs/action.html']

    def __init__(self, token):
        self.token = token
        formdef_type = self.token.context.get('form_type', 'formdef')
        if formdef_type == 'carddef':
            formdef_class = CardDef
        elif formdef_type == 'formdef':
            formdef_class = FormDef
        else:
            raise errors.TraversalError()

        self.formdef = formdef_class.get_by_urlname(self.token.context['form_slug'])
        try:
            self.formdata = self.formdef.data_class().get(self.token.context['form_number_raw'])
        except KeyError:
            raise MissingFormdata()
        self.action = None
        status = self.formdata.get_status()
        for item in status.items:
            if getattr(item, 'identifier', None) == self.token.context['action_id']:
                self.action = item
                break
        else:
            raise MissingOrExpiredToken()

    def _q_index(self):
        template.html_top(title=self.formdef.name)
        form = Form()
        form.add_submit('submit', misc.site_encode(self.token.context['label']))
        if form.is_submitted() and not form.has_errors():
            return self.submit()
        context = {
            'view': self,
            'form': form,
            'html_form': form,
            'message': self.token.context.get('message'),
        }
        return template.QommonTemplateResponse(
            templates=list(self.get_formdef_template_variants(self.templates)), context=context
        )

    def submit(self):
        self.formdata.record_workflow_event('email-button', action_item_id=self.action.id)
        url = jump_and_perform(self.formdata, self.action)
        done_message = self.token.context.get('done_message')
        self.token.remove_self()
        if url:
            return redirect(url)
        context = {
            'view': self,
            'done': True,
            'done_message': done_message,
        }
        return template.QommonTemplateResponse(
            templates=list(self.get_formdef_template_variants(self.templates)), context=context
        )


class GlobalInteractiveActionDirectory(Directory, FormTemplateMixin):
    _q_exports = ['']

    def __init__(self, token):
        self.token = token
        formdef_type = self.token.context.get('form_type', 'formdef')
        if formdef_type == 'carddef':
            formdef_class = CardDef
        elif formdef_type == 'formdef':
            formdef_class = FormDef
        else:
            raise errors.TraversalError()

        self.formdef = formdef_class.get_by_urlname(self.token.context['form_slug'])

        try:
            self.formdata = self.formdef.data_class().get(self.token.context['form_ids'][0])
        except KeyError:
            raise MissingFormdata()

        self.action = None
        for action in self.formdef.workflow.global_actions or []:
            if action.id == self.token.context['action_id']:
                self.action = action
                break
        else:
            raise MissingOrExpiredToken()

    def _q_index(self):
        if get_request().is_in_backoffice():
            if isinstance(self.formdef, CardDef):
                section = 'data_management'
            else:
                section = 'management'
            get_response().breadcrumb.append(('', self.action.name))
            backoffice_html_top(section, title=self.formdef.name)
            template_name = 'wcs/backoffice/global-interactive-action.html'
        else:
            template.html_top(title=self.formdef.name)
            template_name = 'wcs/global-interactive-action.html'

        if len(self.token.context['form_ids']) == 1:
            get_publisher().substitutions.feed(self.formdata)

        form = self.action.get_action_form(self.formdata, user=get_request().user)
        if not form:
            # empty form, nothing to do
            get_session().message = ('error', _('Error: empty action'))
            return redirect(self.token.context['return_url'])
        if not form.is_submitted() or form.has_errors():
            messages = self.action.get_messages()
            context = {
                'html_form': form,
                'action': self.action,
                'ids': self.token.context['form_ids'],
                'formdata': self.formdata,
                'workflow_messages': messages,
            }
            return template.QommonTemplateResponse(templates=[template_name], context=context)

        if len(self.token.context['form_ids']) > 1:
            # mass action
            job = get_response().add_after_job(
                GlobalInteractiveMassActionAfterJob(
                    label=_('Executing task "%s" on forms') % self.action.name,
                    formdef=self.formdef,
                    request_form=get_request().form,
                    user_id=get_request().user.id,
                    action_id=self.action.id,
                    item_ids=self.token.context['form_ids'],
                    return_url=self.token.context['return_url'],
                )
            )
            job.store()
            self.token.remove_self()
            return redirect(job.get_processing_url())

        url = GlobalInteractiveMassActionAfterJob.execute_one(
            get_publisher(), self.formdata, self.action, get_request().user, afterjob=False
        )
        if not url:
            url = self.formdata.get_url(backoffice=bool(get_request().is_in_backoffice()))
        self.token.remove_self()
        return redirect(url)


class GlobalInteractiveMassActionAfterJob(AfterJob):
    def __init__(self, formdef, **kwargs):
        super().__init__(formdef_class=formdef.__class__, formdef_id=formdef.id, **kwargs)

    def execute(self):
        self.total_count = len(self.kwargs['item_ids'])

        # restore request form
        publisher = get_publisher()
        req = HTTPRequest(None, {'SERVER_NAME': publisher.tenant.hostname, 'SCRIPT_NAME': ''})
        req.form = self.kwargs['request_form']

        formdef = self.kwargs['formdef_class'].get(self.kwargs['formdef_id'])
        data_class = formdef.data_class()

        for action in formdef.workflow.global_actions or []:
            if action.id == self.kwargs['action_id']:
                break
        else:
            # maybe action got removed from workflow?
            return

        user = publisher.user_class.get(self.kwargs['user_id'])
        for item_id in self.kwargs['item_ids']:
            publisher._set_request(req)
            formdata = data_class.get(item_id)
            self.execute_one(publisher, formdata, action, user)
            self.increment_count()

    @classmethod
    def execute_one(cls, publisher, formdata, action, user, afterjob=True):
        publisher.reset_formdata_state()
        publisher.substitutions.feed(formdata)

        status = formdata.status
        form = action.get_action_form(formdata, user=user)
        form.method = 'get'
        url = action.handle_form(form, formdata, user=user)
        if afterjob:
            # reset request to avoid emails being created as afterjobs
            publisher._set_request(None)
        if formdata.status == status:
            # if there's no status change run non-interactive items from global action
            formdata.record_workflow_event('global-interactive-action', global_action_id=action.id)
            url = perform_items(action.items, formdata, global_action=True) or url
        else:
            # run actions from new status
            wf_status = formdata.get_status()
            formdata.record_workflow_event('global-interactive-action', global_action_id=action.id)
            url = perform_items(wf_status.items, formdata) or url
        # return url, it is used when the action is performed on a single item
        return url

    def done_action_url(self):
        return self.kwargs['return_url']

    def done_action_label(self):
        return _('Back to Listing')

    def done_button_attributes(self):
        return {'data-redirect-auto': 'true'}
